/**
 * Copyright (C) 2013 Canonical, Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3, as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
 * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors: Tiago Salem Herrmann <tiago.herrmann@canonical.com>
 */

// ofono-qt
#include <ofonomessage.h>

// telepathy-ofono
#include "ofonotextchannel.h"
#include "pendingmessagesmanager.h"

QDBusArgument &operator<<(QDBusArgument&argument, const IncomingAttachmentStruct &attachment)
{
    argument.beginStructure();
    argument << attachment.id << attachment.contentType << attachment.filePath << attachment.offset << attachment.length;
    argument.endStructure();
    return argument;
}

const QDBusArgument &operator>>(const QDBusArgument &argument, IncomingAttachmentStruct &attachment)
{
    argument.beginStructure();
    argument >> attachment.id >> attachment.contentType >> attachment.filePath >> attachment.offset >> attachment.length;
    argument.endStructure();
    return argument;
}


oFonoTextChannel::oFonoTextChannel(oFonoConnection *conn, QStringList phoneNumbers, bool flash, QObject *parent):
    QObject(parent),
    mConnection(conn),
    mPhoneNumbers(phoneNumbers),
    mFlash(flash),
    mMessageCounter(1)
{
    qDBusRegisterMetaType<IncomingAttachmentStruct>();
    qDBusRegisterMetaType<IncomingAttachmentList>();

    Tp::BaseChannelPtr baseChannel;
    if (phoneNumbers.size() == 1) {
        baseChannel = Tp::BaseChannel::create(mConnection,
                                              TP_QT_IFACE_CHANNEL_TYPE_TEXT,
                                              Tp::HandleTypeContact,
                                              mConnection->ensureHandle(mPhoneNumbers[0]));
    } else {
        baseChannel = Tp::BaseChannel::create(mConnection,
                                              TP_QT_IFACE_CHANNEL_TYPE_TEXT);
    }
    Tp::BaseChannelTextTypePtr textType = Tp::BaseChannelTextType::create(baseChannel.data());
    baseChannel->plugInterface(Tp::AbstractChannelInterfacePtr::dynamicCast(textType));

    QStringList supportedContentTypes = QStringList() << "text/plain";

    Tp::UIntList messageTypes = Tp::UIntList() << 
                                Tp::ChannelTextMessageTypeNormal << 
                                Tp::ChannelTextMessageTypeDeliveryReport;
    uint messagePartSupportFlags = 0;
    uint deliveryReportingSupport = Tp::DeliveryReportingSupportFlagReceiveSuccesses;
    mMessagesIface = Tp::BaseChannelMessagesInterface::create(textType.data(),
                                                          supportedContentTypes,
                                                          messageTypes,
                                                          messagePartSupportFlags,
                                                          deliveryReportingSupport);

    mMessagesIface->setSendMessageCallback(Tp::memFun(this,&oFonoTextChannel::sendMessage));

    baseChannel->plugInterface(Tp::AbstractChannelInterfacePtr::dynamicCast(mMessagesIface));

    mGroupIface = Tp::BaseChannelGroupInterface::create(Tp::ChannelGroupFlagCanAdd, conn->selfHandle());
    mGroupIface->setAddMembersCallback(Tp::memFun(this,&oFonoTextChannel::onAddMembers));
    mGroupIface->setRemoveMembersCallback(Tp::memFun(this,&oFonoTextChannel::onRemoveMembers));

    baseChannel->plugInterface(Tp::AbstractChannelInterfacePtr::dynamicCast(mGroupIface));
    addMembers(phoneNumbers);

    mSMSIface = Tp::BaseChannelSMSInterface::create(flash, true);
    baseChannel->plugInterface(Tp::AbstractChannelInterfacePtr::dynamicCast(mSMSIface));

    mBaseChannel = baseChannel;
    mTextChannel = Tp::BaseChannelTextTypePtr::dynamicCast(mBaseChannel->interface(TP_QT_IFACE_CHANNEL_TYPE_TEXT));
    mTextChannel->setMessageAcknowledgedCallback(Tp::memFun(this,&oFonoTextChannel::messageAcknowledged));
    QObject::connect(mBaseChannel.data(), SIGNAL(closed()), this, SLOT(deleteLater()));
}

void oFonoTextChannel::onAddMembers(const Tp::UIntList& handles, const QString& message, Tp::DBusError* error)
{
    addMembers(mConnection->inspectHandles(Tp::HandleTypeContact, handles, error));
}

void oFonoTextChannel::onRemoveMembers(const Tp::UIntList& handles, const QString& message, Tp::DBusError* error)
{
    Q_FOREACH(uint handle, handles) {
        Q_FOREACH(const QString &phoneNumber, mConnection->inspectHandles(Tp::HandleTypeContact, Tp::UIntList() << handle, error)) {
            mPhoneNumbers.removeAll(phoneNumber);
        }
        mMembers.removeAll(handle);
    }
    mGroupIface->removeMembers(handles);
}

void oFonoTextChannel::addMembers(QStringList phoneNumbers)
{
    Tp::UIntList handles;
    Q_FOREACH(const QString &phoneNumber, phoneNumbers) {
        uint handle = mConnection->ensureHandle(phoneNumber);
        handles << handle;
        if (!mPhoneNumbers.contains(phoneNumber)) {
            mPhoneNumbers << phoneNumber;
        }
        if (!mMembers.contains(handle)) {
            mMembers << handle;
        }
    }
    mGroupIface->addMembers(handles, phoneNumbers);
}

Tp::UIntList oFonoTextChannel::members()
{
    return mMembers;
}

oFonoTextChannel::~oFonoTextChannel()
{
    Q_FOREACH(const QStringList &fileList, mFilesToRemove) {
        Q_FOREACH(const QString& file, fileList) {
            QFile::remove(file);
        }
    }
}

Tp::BaseChannelPtr oFonoTextChannel::baseChannel()
{
    return mBaseChannel;
}

void oFonoTextChannel::sendDeliveryReport(const QString &messageId, uint handle, Tp::DeliveryStatus status)
{
    Tp::MessagePartList partList;
    Tp::MessagePart header;
    header["message-sender"] = QDBusVariant(handle);
    header["message-sender-id"] = QDBusVariant(mPhoneNumbers[0]);
    header["message-type"] = QDBusVariant(Tp::ChannelTextMessageTypeDeliveryReport);
    header["delivery-status"] = QDBusVariant(status);
    header["delivery-token"] = QDBusVariant(messageId);
    partList << header;
    mTextChannel->addReceivedMessage(partList);
}

void oFonoTextChannel::deliveryReportReceived(const QString &messageId, uint handle, bool success)
{
    sendDeliveryReport(messageId, handle, success ? Tp::DeliveryStatusDelivered : Tp::DeliveryStatusPermanentlyFailed);
}

void oFonoTextChannel::messageAcknowledged(const QString &id)
{
    Q_EMIT messageRead(id);
}

QString oFonoTextChannel::sendMessage(Tp::MessagePartList message, uint flags, Tp::DBusError* error)
{
    bool success = true;
    Tp::MessagePart header = message.at(0);
    Tp::MessagePart body = message.at(1);
    QString objpath;

    bool mms = header["mms"].variant().toBool();

    if (mms) {
        // pop header out
        message.removeFirst();
        OutgoingAttachmentList attachments;
        // FIXME group chat
        QString phoneNumber = mPhoneNumbers[0];
        uint handle = mConnection->ensureHandle(phoneNumber);
        QStringList temporaryFiles;
        Q_FOREACH(const Tp::MessagePart &part, message) {
            OutgoingAttachmentStruct attachment;
            attachment.id = part["identifier"].variant().toString();
            attachment.contentType = part["content-type"].variant().toString();
            QString attachmentsPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + "/telepathy-ofono/attachments";
            if (!QDir().exists(attachmentsPath) && !QDir().mkpath(attachmentsPath)) {
                qCritical() << "Failed to create attachments directory";
                objpath = QDateTime::currentDateTimeUtc().toString(Qt::ISODate) + "-" + QString::number(mMessageCounter++);
                error->set(TP_QT_ERROR_INVALID_ARGUMENT, "Failed to create attachments to disk");
                mPendingDeliveryReportPermanentlyFailed[objpath] = handle;
                QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
                return objpath;
            }
            QTemporaryFile file(attachmentsPath + "/attachmentXXXXXX");
            file.setAutoRemove(false);
            if (!file.open()) {
                Q_FOREACH(const QString& file, temporaryFiles) {
                    QFile::remove(file);
                }
                objpath = QDateTime::currentDateTimeUtc().toString(Qt::ISODate) + "-" + QString::number(mMessageCounter++);
                error->set(TP_QT_ERROR_INVALID_ARGUMENT, "Failed to create attachments to disk");
                mPendingDeliveryReportPermanentlyFailed[objpath] = handle;
                QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
                return objpath;
            }
            temporaryFiles << file.fileName();
            file.write(part["content"].variant().toByteArray());
            file.close();
            attachment.filePath = file.fileName();
            attachments << attachment;
        }
        objpath = mConnection->sendMMS(mPhoneNumbers, attachments).path();
        if (objpath.isEmpty()) {
            Q_FOREACH(const QString& file, temporaryFiles) {
                QFile::remove(file);
            }
            // give a temporary id for this message
            objpath = QDateTime::currentDateTimeUtc().toString(Qt::ISODate) + "-" + QString::number(mMessageCounter++);
            // TODO: get error message from nuntium
            error->set(TP_QT_ERROR_INVALID_ARGUMENT, "Failed to send MMS");
            mPendingDeliveryReportPermanentlyFailed[objpath] = handle;
            QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
            return objpath;
        }
        MMSDMessage *msg = new MMSDMessage(objpath, QVariantMap(), this);
        QObject::connect(msg, SIGNAL(propertyChanged(QString,QVariant)), SLOT(onMMSPropertyChanged(QString,QVariant)));
        mPendingDeliveryReportUnknown[objpath] = handle;
        QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
        if (temporaryFiles.size() > 0) {
            mFilesToRemove[objpath] = temporaryFiles;
        }
        return objpath;
    }

    if (mPhoneNumbers.size() == 1) {
        QString phoneNumber = mPhoneNumbers[0];
        uint handle = mConnection->ensureHandle(phoneNumber);
        objpath = mConnection->messageManager()->sendMessage(phoneNumber, body["content"].variant().toString(), success).path();
        if (objpath.isEmpty() || !success) {
            if (!success) {
                qWarning() << mConnection->messageManager()->errorName() << mConnection->messageManager()->errorMessage();
            } else {
                error->set(TP_QT_ERROR_INVALID_ARGUMENT, mConnection->messageManager()->errorMessage());
            }
            // create an unique id for this message that ofono failed to send
            objpath = QDateTime::currentDateTimeUtc().toString(Qt::ISODate) + "-" + QString::number(mMessageCounter++);
            mPendingDeliveryReportPermanentlyFailed[objpath] = handle;
            QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
            return objpath;
        }
        OfonoMessage *msg = new OfonoMessage(objpath);
        if (msg->state() == "") {
            // message was already sent or failed too fast (this case is only reproducible with the emulator)
            msg->deleteLater();
            mPendingDeliveryReportUnknown[objpath] = handle;
            QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
            return objpath;
        }
        // FIXME: track pending messages only if delivery reports are enabled. We need a system config option for it.
        PendingMessagesManager::instance()->addPendingMessage(objpath, mPhoneNumbers[0]);
        QObject::connect(msg, SIGNAL(stateChanged(QString)), SLOT(onOfonoMessageStateChanged(QString)));
        return objpath;
    } else {
        bool someMessageSent = false;
        QString lastPhoneNumber;
        Q_FOREACH(const QString &phoneNumber, mPhoneNumbers) {
            uint handle = mConnection->ensureHandle(mPhoneNumbers[0]);
            objpath = mConnection->messageManager()->sendMessage(phoneNumber, body["content"].variant().toString(), success).path();
            lastPhoneNumber = phoneNumber;
            // dont fail if this is a group chat as we cannot track individual messages
            if (objpath.isEmpty() || !success) {
                if (!success) {
                    qWarning() << mConnection->messageManager()->errorName() << mConnection->messageManager()->errorMessage();
                } else {
                    error->set(TP_QT_ERROR_INVALID_ARGUMENT, mConnection->messageManager()->errorMessage());
                }
                continue;
            }
            someMessageSent = true;
        }
        if (!someMessageSent) {
            // for group chat we only fail if all the messages failed to send
            objpath = QDateTime::currentDateTimeUtc().toString(Qt::ISODate) + "-" + QString::number(mMessageCounter++);
            uint handle = mConnection->ensureHandle(lastPhoneNumber);
            mPendingDeliveryReportPermanentlyFailed[objpath] = handle;
            QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
            return objpath;
        }
        OfonoMessage *msg = new OfonoMessage(objpath);
        if (msg->state() == "") {
            // message was already sent or failed too fast (this case is only reproducible with the emulator)
            msg->deleteLater();
            uint handle = mConnection->ensureHandle(lastPhoneNumber);
            mPendingDeliveryReportUnknown[objpath] = handle;
            QTimer::singleShot(0, this, SLOT(onProcessPendingDeliveryReport()));
            // return only the last one in case of group chat for history purposes
            return objpath;
        }
        QObject::connect(msg, SIGNAL(stateChanged(QString)), SLOT(onOfonoMessageStateChanged(QString)));
        return objpath;
    }
}

void oFonoTextChannel::onMMSPropertyChanged(QString property, QVariant value)
{
    qDebug() << "oFonoTextChannel::onMMSPropertyChanged" << property << value;
    MMSDMessage *msg = qobject_cast<MMSDMessage*>(sender());
    // FIXME - mms groupchat
    uint handle = mConnection->ensureHandle(mPhoneNumbers[0]);
    if (!msg) {
        return;
    }
    if (property == "Status") {
        Tp::DeliveryStatus status = Tp::DeliveryStatusUnknown;
        if (value == "Sent") {
            status = Tp::DeliveryStatusAccepted;
        } else if (value == "TransientError" || value == "PermanentError") {
            // transient error in telepathy means it is still trying, so we need to
            // set a permanent error here so the user can retry
            status = Tp::DeliveryStatusPermanentlyFailed;
        } else if (value == "draft") {
            // while it is draft we dont actually send a delivery report
            return;
        }
        Q_FOREACH(const QString& file, mFilesToRemove[msg->path()]) {
            QFile::remove(file);
        }
        mFilesToRemove.remove(msg->path());
        sendDeliveryReport(msg->path(), handle, status);
    }
}

void oFonoTextChannel::onProcessPendingDeliveryReport()
{
    QMapIterator<QString, uint> iterator(mPendingDeliveryReportPermanentlyFailed);
    while(iterator.hasNext()) {
        iterator.next();
        sendDeliveryReport(iterator.key(), iterator.value(), Tp::DeliveryStatusPermanentlyFailed);
    }
    iterator = mPendingDeliveryReportDelivered;
    while(iterator.hasNext()) {
        iterator.next();
        sendDeliveryReport(iterator.key(), iterator.value(), Tp::DeliveryStatusDelivered);
    }
    iterator = mPendingDeliveryReportAccepted;
    while(iterator.hasNext()) {
        iterator.next();
        sendDeliveryReport(iterator.key(), iterator.value(), Tp::DeliveryStatusAccepted);
    }
    iterator = mPendingDeliveryReportTemporarilyFailed;
    while(iterator.hasNext()) {
        iterator.next();
        sendDeliveryReport(iterator.key(), iterator.value(), Tp::DeliveryStatusTemporarilyFailed);
    }
    iterator = mPendingDeliveryReportUnknown;
    while(iterator.hasNext()) {
        iterator.next();
        sendDeliveryReport(iterator.key(), iterator.value(), Tp::DeliveryStatusUnknown);
    }

    mPendingDeliveryReportTemporarilyFailed.clear();
    mPendingDeliveryReportPermanentlyFailed.clear();
    mPendingDeliveryReportDelivered.clear();
    mPendingDeliveryReportAccepted.clear();
    mPendingDeliveryReportUnknown.clear();
}

void oFonoTextChannel::onOfonoMessageStateChanged(QString status)
{
    OfonoMessage *msg = static_cast<OfonoMessage *>(sender());
    if(msg) {
        Tp::DeliveryStatus delivery_status;
        if (status == "sent") {
            delivery_status = Tp::DeliveryStatusAccepted;
            msg->deleteLater();
        } else if(status == "failed") {
            delivery_status = Tp::DeliveryStatusPermanentlyFailed;
            PendingMessagesManager::instance()->removePendingMessage(msg->path());
            msg->deleteLater();
        } else if(status == "pending") {
            delivery_status = Tp::DeliveryStatusTemporarilyFailed;
        } else {
            delivery_status = Tp::DeliveryStatusUnknown;
        }

        sendDeliveryReport(msg->path(), mConnection->ensureHandle(mPhoneNumbers[0]), delivery_status);
    }
}

void oFonoTextChannel::messageReceived(const QString &message, uint handle, const QVariantMap &info)
{
    Tp::MessagePartList partList;

    Tp::MessagePart body;
    body["content-type"] = QDBusVariant("text/plain");
    body["content"] = QDBusVariant(message);

    Tp::MessagePart header;
    header["message-token"] = QDBusVariant(info["SentTime"].toString() +"-" + QString::number(mMessageCounter++));
    header["message-received"] = QDBusVariant(QDateTime::currentDateTime().toTime_t());
    header["message-sender"] = QDBusVariant(handle);
    header["message-sender-id"] = QDBusVariant(mPhoneNumbers[0]);
    header["message-type"] = QDBusVariant(Tp::ChannelTextMessageTypeNormal);
    partList << header << body;

    mTextChannel->addReceivedMessage(partList);
}

void oFonoTextChannel::mmsReceived(const QString &id, uint handle, const QVariantMap &properties)
{
    Tp::MessagePartList message;
    QString subject = properties["Subject"].toString();
    QString smil = properties["Smil"].toString();

    Tp::MessagePart header;
    header["message-token"] = QDBusVariant(id);
    header["message-sender"] = QDBusVariant(handle);
    header["message-received"] = QDBusVariant(QDateTime::currentDateTimeUtc().toTime_t());
    header["message-type"] = QDBusVariant(Tp::DeliveryStatusDelivered);
    header["mms"] = QDBusVariant(true);
    if (!subject.isEmpty())
    {
        header["subject"] = QDBusVariant(subject);
    }
    message << header;
    IncomingAttachmentList mmsdAttachments = qdbus_cast<IncomingAttachmentList>(properties["Attachments"]);
    Q_FOREACH(const IncomingAttachmentStruct &attachment, mmsdAttachments) {
        QFile attachmentFile(attachment.filePath);
        if (!attachmentFile.open(QIODevice::ReadOnly)) {
            qWarning() << "fail to load attachment" << attachmentFile.errorString() << attachment.filePath;
            continue;
        }
        // FIXME check if we managed to read the total attachment file
        attachmentFile.seek(attachment.offset);
        QByteArray fileData = attachmentFile.read(attachment.length);
        Tp::MessagePart part;
        part["content-type"] =  QDBusVariant(attachment.contentType);
        part["identifier"] = QDBusVariant(attachment.id);
        part["content"] = QDBusVariant(fileData);
        part["size"] = QDBusVariant(attachment.length);

        message << part;
    }

    if (!smil.isEmpty()) {
        Tp::MessagePart part;
        part["content-type"] =  QDBusVariant(QString("application/smil"));
        part["identifier"] = QDBusVariant(QString("smil"));
        part["content"] = QDBusVariant(smil);
        part["size"] = QDBusVariant(smil.size());
        message << part;
    }

    mTextChannel->addReceivedMessage(message);
}
